library("scales")
library("tidyverse")

LINE_SIZE <- 0.3

DATE_LIMITS <- as.Date(c(
	"2022-09-18",
	"2023-01-16"
))

WANTED_COUNTRIES <- c(
	"ir"
)

WANTED_TRANSPORTS <- c(
	"total",
	"<OR> relay",
	"<OR> bridge",
	"meek",
	"snowflake",
	"obfs4"
)

BRIDGE_TRANSPORTS <- c(
	"obfs4",
	"snowflake",
	"meek",
	"<OR> bridge"
)

COLOR_PALETTE <- c(
	"total" = "gray20",
	"<OR> relay" = "gray20",
	"obfs4" = "seagreen",
	"snowflake" = "dodgerblue",
	"meek" = "darkorange",
	"<OR> bridge" = "gray30"
)

FILL_PALETTE <- COLOR_PALETTE
FILL_PALETTE[["total"]] = "transparent"
FILL_PALETTE[["<OR> relay"]] = "transparent"

LINETYPE_PALETTE <- c(
	"total" = "dotted",
	"<OR> relay" = "solid",
	"obfs4" = "solid",
	"snowflake" = "solid",
	"meek" = "solid",
	"<OR> bridge" = "solid"
)

missing_dates_fill_na <- function(tb, ...) {
	# right_join(tb, tibble(date = seq.Date(min(tb$date), max(tb$date), "days")))
	# complete(tb, date = seq.Date(min(tb$date), max(tb$date), "days"), nesting(country, transport))
	complete(tb, date = seq.Date(min(tb$date), max(tb$date), "days"), ...)
}

# Returns a vector z that minimizes sum((z - y)^2), subject to the conditions
# that all elements of z are at least M, and the difference between any two
# elements of z is at least S.
place_no_overlap <- function(y, M, S) {
	# We deal with the input in reverse-sorted order, then undo the
	# permutation in the result vector when finished.
	o <- order(y, decreasing=TRUE)
	y <- y[o]
	# Our goal is to divide y into contiguous, noninteracting clumps,
	# adjacent elements of which are separated by exactly S. Then, we place
	# each clump to minimize its individual cost.
	#
	# We look at clumps of every possible size starting at the top of y,
	# and break off the clump whose highest point is the greatest, when the
	# clump is placed optimally.
	z <- c()
	while (length(y) > 0) {
		# The cost of placing a clump y[1:i], with the highest element
		# of the clump placed at point zp, is
		#   sum((zp - (0:(i-1))*S - y[1:i])^2).
		# (The highest element is placed as zp, the next highest at
		# zp-1*S, then zp-2*S, etc.) The derivative of this function is
		#   2*sum(zp - (0:(i-1))*S - y[1:i]) = 2*(i*zp - sum((0:(i-1))*S + y[i:1])),
		# which is equal to zero when
		#   zp = sum((0:(i-1))*S + y[i:1]) / i.
		zp <- cumsum((0:(length(y)-1))*S + y) / (1:length(y))
		# We must additionally enforce the M condition, which means
		# that the highest element of a clump of size i cannot be
		# placed lower than M + (i-1)*S.
		zp <- pmax(zp, M + (0:(length(y)-1))*S)
		# Find the clump which would be placed the highest.
		i <- which.max(zp)
		# Break off the top i elements as a clump and place them at zp,
		# zp-1*S, zp-2*S, etc.
		z <- c(z, zp[[i]] - (0:(i-1))*S)
		if (i < length(y))
			y <- y[(i+1):length(y)]
		else
			y <- c()
	}
	z[order(o)]
}

(function() {
	args <- commandArgs(trailingOnly = TRUE)
	if (length(args) != 3) {
		stop("usage: Rscript userstats-ir.r userstats-relay-country-....csv userstats-bridge-combined-....csv snowflake-userstats-bridge-combined.csv")
	}
	relay_country_csv_path <<- args[[1]]
	bridge_combined_csv_path <<- args[[2]]
	snowflake_override_csv_path <<- args[[3]]
})()

relay_country <- read_csv(relay_country_csv_path, comment = "#") %>%
	group_by(country) %>% missing_dates_fill_na() %>% ungroup()
bridge_combined <- read_csv(bridge_combined_csv_path, comment = "#") %>%
	filter(transport != "snowflake") %>%
	group_by(country, transport) %>% missing_dates_fill_na(country, transport) %>% ungroup()
snowflake_override <- read_csv(snowflake_override_csv_path, comment = "#") %>%
	filter(country %in% WANTED_COUNTRIES) %>%
	group_by(country, transport) %>% missing_dates_fill_na(country, transport) %>% ungroup()

together <- bind_rows(
	relay_country %>%
		mutate(transport = "<OR> relay", low = users, high = users),
	bridge_combined %>%
		mutate(transport = ifelse(transport == "<OR>", "<OR> bridge", transport)),
	snowflake_override
)

together <- bind_rows(
	together,
	together %>%
		group_by(date) %>%
		summarize(low = sum(low), high = sum(high)) %>%
		mutate(transport = "total", frac = 100)
)

# Filter out low-frequency transports (e.g. scramblesuit, obfs3) (but include
# them in the total).
together <- filter(together, transport %in% WANTED_TRANSPORTS)

# Throw away measurements outside our date range. (Do it this way, instead of
# setting limits on scale_x_date, in order to draw labels at points outside the
# date range.)
together <- filter(together, DATE_LIMITS[[1]] <= date & date <= DATE_LIMITS[[2]])

p <- ggplot(together) +
	geom_ribbon(aes(x = date, ymin = low, ymax = high, fill = transport), alpha = 0.5) +
	geom_line(aes(x = date, y = low, color = transport, linetype = transport), size = LINE_SIZE, alpha = 0.9) +
	geom_line(aes(x = date, y = high, color = transport, linetype = transport), size = LINE_SIZE, alpha = 0.9) +
	geom_point(aes(x = date, y = (low + high) / 2, color = transport), size = LINE_SIZE, alpha = 0.9) +
	# Draw the line labels in the right margin. theme(plot.margin) below
	# makes room for this.
	geom_text(data = together %>%
		filter(date <= DATE_LIMITS[[2]] & !is.na(low)) %>%
		group_by(transport) %>%
		arrange(desc(date)) %>%
		slice(1) %>%
		ungroup() %>%
		mutate(date = max(date)) %>%
		mutate(y = place_no_overlap((low + high) / 2, 0, 4000)),
		aes(x = date, y = y, group = transport, label = transport, color = transport),
		hjust = 0, position = position_nudge(x = 1.0),
		size = 2) +
	# scale_x_date(date_breaks = "1 week", minor_breaks = NULL, date_minor_breaks = waiver(), labels = function(x) ifelse(as.POSIXlt(x)$mday == 1, strftime(x, "%d\n%b"), strftime(x, "%d"))) +
	scale_y_continuous(breaks = 50000*0:10, label = comma_format()) +
	scale_color_manual(breaks = names(COLOR_PALETTE), values = COLOR_PALETTE) +
	scale_fill_manual(breaks = names(COLOR_PALETTE), values = FILL_PALETTE) +
	scale_linetype_manual(breaks = names(COLOR_PALETTE), values = LINETYPE_PALETTE) +
	coord_cartesian(xlim = DATE_LIMITS, ylim = c(0, max(together$high, na.rm = TRUE)), expand = FALSE, clip = "off") +
	guides(
		color = "none",
		fill = "none",
		linetype = "none"
	) +
	labs(title = "Tor users from Iran, September 2022–January 2023", x = NULL, y = "average simultaneous users") +
	theme_minimal() +
	theme(plot.margin = unit(c(0.1, 2.0, 0.1, 0.1), "cm"))
ggsave("tor-userstats-ir.png", p, width = 6, height = 4, dpi = 300)
